/* spawn.c - keeps respawning a process when it exits
 * Time-stamp: <2004-01-27 03:36:31 njk>
 *  
 * (C) 2004 Nicholas J. Kain <njk@aerifal.cx>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "signals.h"

#define MAX_PATH_LENGTH 512			/* Maximum filename+path length */

#define THROTTLE 1					/* Throttle rapidly respawning jobs */
#define MIN_FORK_SEC 10				/* Minimum time required between exec()s */

static void sighandler(int sig) {
	switch (sig) {
		case SIGTERM:
		case SIGINT:
			/* exit, but leave our child process running */
			exit(EXIT_SUCCESS);
			break;
		case SIGCHLD:
			while (waitpid(0, NULL, WNOHANG) >= 0)
			break;
	}
}

static void fix_signals(void) {
  disable_signal(SIGPIPE);
  disable_signal(SIGUSR1);
  disable_signal(SIGUSR2);
  disable_signal(SIGTSTP);
  disable_signal(SIGTTIN);
  disable_signal(SIGHUP);
 
  hook_signal(SIGCHLD, sighandler, 0);
  hook_signal(SIGINT, sighandler, 0);
  hook_signal(SIGTERM, sighandler, 0);
}

static void open_console(void) {
	int fd;
	if ( (fd = open("/dev/console", O_RDWR|O_NOCTTY)) >=0) {
		dup2(fd, 0);
		dup2(fd, 1);
		dup2(fd, 2);
		if (fd>2) close(fd);
	}
}

static void execute(int argc, char **argv) {
	char *av[argc-1], *p;
	struct stat stat_buf;
	int i;

	if (argc < 1) exit(EXIT_FAILURE);
	if (argv == NULL) exit(EXIT_FAILURE);
	if (stat(argv[1], &stat_buf) == -1) exit(EXIT_FAILURE);
	
	p = strrchr(argv[1], '/');
	if (p != NULL) {
		av[0] = ++p;
	} else {
		av[0] = argv[1];
	}
	
	for (i=2; i<= argc; i++) {
		av[i-1] = argv[i];
	}
	
	execvp(argv[1], av);
}

static void do_exec(int argc, char **args) {
	pid_t t, r;
	if (!args) return;
	
	switch ((t = fork())) {
		case 0:
			setsid();
			open_console();
		//	tcsetpgrp(0, getpgrp());
			execute(argc, args);
			break;
		case -1:
			perror("fork");
			exit(EXIT_FAILURE);
			break;
		default:
			while (1) {
				r = waitpid(t, NULL, 0);
				if (r == t) break;
				if (r == -1 && errno != EINTR) break;
			}
			break;
	}
}

static void do_spawn(int argc, char **args) {
	time_t runtime, curtime, t;
	if (!args) return;
	for(;;) {
#ifdef THROTTLE
		runtime = time(0);
#endif
		do_exec(argc, args);
#ifdef THROTTLE
		curtime = time(0);
		if (runtime > curtime) {
			/* Horrible clock slew or time overflow, expect
			 * the worst and die. */
			break;
		}
		if ((t = curtime - runtime) < MIN_FORK_SEC) {
			/* Throttle respawn of broken instant-exit tasks */
			sleep(MIN_FORK_SEC - t);
		}
#endif
	}
}

int main(int argc, char** argv) {
	int c, fd;

	fix_signals();

	switch(fork()) {
		case 0:
			if (setsid() == -1) {
				perror("setsid");
				exit(EXIT_FAILURE);
			}
			for (c=0; c<3; c++) {
				close(c);
				fd = open("/dev/null", O_RDWR);
				if (fd != c) {
					perror("open");
					exit(EXIT_FAILURE);
				}
			}
			break;
		case -1:
			perror("fork");
			exit(EXIT_FAILURE);
			break;
		default:
			exit(EXIT_SUCCESS);
	}
	
	do_spawn(argc, argv);
  
	exit(EXIT_FAILURE);
}

